def chained_join(*relationships):
    """
    Return a chained Join object for given relationships.
    """
    property_ = relationships[0].property

    if property_.secondary is not None:
        from_ = property_.secondary.join(
            property_.mapper.class_.__table__, property_.secondaryjoin
        )
    else:
        from_ = property_.mapper.class_.__table__
    for relationship in relationships[1:]:
        prop = relationship.property
        if prop.secondary is not None:
            from_ = from_.join(prop.secondary, prop.primaryjoin)

            from_ = from_.join(prop.mapper.class_, prop.secondaryjoin)
        else:
            from_ = from_.join(prop.mapper.class_, prop.primaryjoin)
    return from_



import re
from importlib.metadata import metadata


def get_sqlalchemy_version(version=metadata('sqlalchemy')['Version']):
    """Extract the sqlalchemy version as a tuple of integers."""

    match = re.search(r'^(\d+)(?:\.(\d+)(?:\.(\d+))?)?', version)
    try:
        return tuple(int(v) for v in match.groups() if v is not None)
    except AttributeError:
        return ()



from sqlalchemy import types

from ..scalar_coercible import ScalarCoercible
from .pendulum_date import PendulumDate


class EnrichedDateType(types.TypeDecorator, ScalarCoercible):
    """
    Supported for pendulum only.

    Example::


        from sqlalchemy_utils import EnrichedDateType
        import pendulum


        class User(Base):
            __tablename__ = 'user'
            id = sa.Column(sa.Integer, primary_key=True)
            birthday = sa.Column(EnrichedDateType(type="pendulum"))


        user = User()
        user.birthday = pendulum.datetime(year=1995, month=7, day=11)
        session.add(user)
        session.commit()
    """

    impl = types.Date
    cache_ok = True

    def __init__(self, date_processor=PendulumDate, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.date_object = date_processor()

    def _coerce(self, value):
        return self.date_object._coerce(self.impl, value)

    def process_bind_param(self, value, dialect):
        return self.date_object.process_bind_param(self.impl, value, dialect)

    def process_result_value(self, value, dialect):
        return self.date_object.process_result_value(self.impl, value, dialect)

    def process_literal_param(self, value, dialect):
        return value

    @property
    def python_type(self):
        return self.impl.type.python_type



from functools import total_ordering

from .. import i18n
from ..utils import str_coercible


@str_coercible
@total_ordering
class WeekDay:
    NUM_WEEK_DAYS = 7

    def __init__(self, index):
        if not (0 <= index < self.NUM_WEEK_DAYS):
            raise ValueError('index must be between 0 and %d' % self.NUM_WEEK_DAYS)
        self.index = index

    def __eq__(self, other):
        if isinstance(other, WeekDay):
            return self.index == other.index
        else:
            return NotImplemented

    def __hash__(self):
        return hash(self.index)

    def __lt__(self, other):
        return self.position < other.position

    def __repr__(self):
        return f'{self.__class__.__name__}({self.index!r})'

    def __unicode__(self):
        return self.name

    def get_name(self, width='wide', context='format'):
        names = i18n.babel.dates.get_day_names(width, context, i18n.get_locale())
        return names[self.index]

    @property
    def name(self):
        return self.get_name()

    @property
    def position(self):
        return (self.index - i18n.get_locale().first_week_day) % self.NUM_WEEK_DAYS
